math.randomseed(os.time())
-- a 128x128 map, so every field is 4x4 pixels
-- at first, we have a 4x4 map, whose size will be iteratively increased
-- field types:
--  'w' water
--  'l' unclaimed land
--   1  first empire
--   2  second empire
--   3  third empire
--   etc.

--require 'ser'

local FIELDS_PER_PROVINCE = 100

local map = {}
local pmap = {}
local provinces = {}
local max = 128
local pix = 4
local prec = 1--1/4
local neighbours = {}
local istep = max/pix
local generatedempires
local empires = {}
local total_land = {}
local empirespaces = {}
local changed_pixels = {}
local fb = love.graphics.newFramebuffer(512,512)
local altfb = love.graphics.newFramebuffer(512,512)
love.graphics.setRenderTarget(altfb)
love.graphics.setRenderTarget()
local emp_fbs = {}
local province_fbs = {}
local province_image = {}
local number_of_provinces = {}
local mode = 0
local total_pix = 0
local pix_so_far = 0
local colorname = {
	'red', 'yellow', 'pink', 'grey', 'black',
	'white', 'navy', '...', '...', '...', 'dark green', '...', '...'
}

local mr = math.random
local function choose(tbl)
	return tbl[mr(#tbl)]
end

local maxemp = 0
local colors = {w = {100, 100, 255}, l = {100, 255, 100}, {255, 0, 0},
		{255, 255, 0}, {255, 0, 255}, {128, 128, 128}, {0, 0, 0},
		{255, 255, 255}, {0, 255, 255}, {128, 0, 255}, {128, 255, 0},
		{255, 128, 255}, {0, 128, 0}, {255, 128, 0}, {0, 128, 255}}
local xposses = {0, 1, 1, 0}
local yposses = {0, 0, 1, 1}
local xdir = {-prec, 0, prec, 0}--{-.5, 0, .5, 0}
local ydir = {0, -prec, 0, prec}--{0, -.5, 0, .5}

--print_r(map)
function love.update2(dt)
	if not init then
		init = true
		love.graphics.setCaption("AStrat // Phase 0: generating land")
		return
	end
	if istep > prec then
		genmap()
		if istep <= prec then
			love.graphics.setCaption("AStrat // Phase 1: generating empires")
			for k,v in pairs(map) do
				pmap[k] = {}
				for kk,_ in pairs(v) do
					pmap[k][kk] = false
				end
			end
		end
		love.timer.sleep(400)
	elseif not generatedempires then
		genempires()
	elseif not landfull then
		stepconquer()
	elseif not provincesfull then
		stepprovince()
	elseif not completevictory then
		conquer()
	else
		love.timer.sleep(200)
	end
	love.timer.sleep(10)
end

local et = {}
local function getenv(i,j)
	return {(map[i-istep] or et)[j], (map[i] or et)[j-istep], (map[i+istep] or et)[j], (map[i] or et)[j+istep]}
end

function genmap()
	local nmap = {}
	local hstep = istep*.5
	for i = 1,max,istep do
		nmap[i] = {}
		nmap[i+hstep] = {}
		for j = 1,max,istep do
			local t = map[i][j]
			local env = getenv(i,j)
			local td = map[i][j+1]
			local tu = map[i][j-1]
			local tl = (map[i+1] or {})[j]
			local tr = (map[i-1] or {})[j]
			for rot = 1, 4 do
				local temp = {t,t,t}
				if t == 'l' then table.insert(temp, t) end
				if env[rot] then table.insert(temp, env[rot]) end
				if env[rot%4+1] then table.insert(temp, env[rot%4+1]) end
				nmap[i+hstep*xposses[rot]][j+hstep*yposses[rot]] = choose(temp)
			end
		end
	end
	map = nmap
	istep = hstep
end

function genempires()
	maxemp = maxemp + 1
	emp_fbs[maxemp] = love.graphics.newFramebuffer(512,512)
	love.graphics.setRenderTarget(emp_fbs[maxemp])
	love.graphics.setRenderTarget()
	total_pix = total_pix + 1
	while true do
		local x,y = mr(max), mr(max)
		if map[x][y] == 'l' then
			map[x][y] = maxemp
			table.insert(empires, maxemp)
			empirespaces[maxemp] = {{x,y}}
			changed_pixels[#changed_pixels+1] = empirespaces[maxemp][1]
			total_land[maxemp] = 1
			break
		end
	end
	if maxemp > 5 then
		generatedempires = true
		love.graphics.setCaption("AStrat // Phase 2: growing empires")
	end
end
--envs = {l = 0, w = 0, 0, 0, 0, 0, 0, 0}
function stepconquer()
	--print('l:'..envs.l, 'w:'..envs.w, '1:'..envs[1], '2:'..envs[2], '3:'..envs[3],
	--	'4:'..envs[4], '5:'..envs[5], '6:'..envs[6])
	for I=1,#empires do
		local c = empires[I]
		if c then
		if #empirespaces[c] == 0 then
			table.remove(empires, I)
			--print('empire '..c..' T')
		else
			local m_i = mr(#empirespaces[c])
			local x,y = unpack(empirespaces[c][m_i])
			local env = getenv(x,y)
			--local own_i = 0
			local avail = {}
			for i=1,4 do
--				if env[i] then envs[env[i]] = envs[env[i]] + 1 end
				--if env[i] == c then
				--	own_i = own_i + 1
				if env[i] == 'l' --[[or type(env[i]) == 'number']] then
					table.insert(avail, {xdir[i]+x, ydir[i]+y})
				end
			end
			--print(#avail)
			--if own_i == 4 then
			--	OI_C = OI_C + 1
			--	table.remove(empirespaces[c], m_i)
			--end
			if #avail > 0 then
				local L = choose(avail)
				local X,Y = unpack(L)
				table.insert(empirespaces[c], L)
				changed_pixels[#changed_pixels+1] = L
				--print(unpack(empirespaces[c]))
				map[X][Y] = c
				total_land[c] = total_land[c] + 1
				total_pix = total_pix + 1
			else
				table.remove(empirespaces[c], m_i)
			end
		end
		end
	end
	if #empires == 0 then
		-- next stage
		landfull = true
		love.graphics.setCaption("AStrat // Phase 3: creating provinces")
		local total_j = 0
		for i=1,maxemp do
			local maxp = math.ceil(total_land[i] / FIELDS_PER_PROVINCE)
			--print(total_land[i], maxp)
			number_of_provinces[i] = maxp
			pix_so_far = pix_so_far + maxp
			for j=1,maxp do
				while true do
					local x,y = mr(max), mr(max)
					if map[x][y] == i and not pmap[x][y] then
						local INDEX = j + total_j
						province_fbs[INDEX] = love.image.newImageData(512,512)
						province_image[INDEX] = love.graphics.newImage(province_fbs[INDEX])
						provinces[INDEX] = i
						neighbours[INDEX] = {}
						pmap[x][y] = INDEX
						table.insert(empires, INDEX)
						empirespaces[INDEX] = {{x,y}}
						changed_pixels[#changed_pixels+1] = empirespaces[INDEX][1]
						break
					end
				end
			end
			total_j = total_j + maxp
		end
		love.graphics.setCaption("AStrat // Phase 4: filling provinces")
	end
end

function stepprovince()
	for I=1,#empires do
		local c = empires[I]
		if c then
		if #empirespaces[c] == 0 then
			table.remove(empires, I)
		else
			local m_i = mr(#empirespaces[c])
			local x,y = unpack(empirespaces[c][m_i])
			local env = getenv(x,y)
			local avail = {}
			for i=1,4 do
				local prov = pmap[xdir[i]+x][ydir[i]+y]
				if prov then
					--enter neighbour
					neighbours[prov][c] = true
					neighbours[c][prov] = true
				elseif env[i] == provinces[c] then
					table.insert(avail, {xdir[i]+x, ydir[i]+y})
				end
			end
			if #avail > 0 then
				local L = choose(avail)
				local X,Y = unpack(L)
				table.insert(empirespaces[c], L)
				changed_pixels[#changed_pixels+1] = L
				pmap[X][Y] = c
				pix_so_far = pix_so_far + 1
			else
				table.remove(empirespaces[c], m_i)
			end
		end
		end
	end
	if #empires == 0 then
		provincesfull = true
		love.graphics.setCaption("AStrat // Phase 5: finding neighbours")
		for i, val in ipairs(neighbours) do -- converting sets to lists for maximal performance
			local n = {}
			for k,_ in pairs(val) do
				n[#n+1] = k
			end
			neighbours[i] = n
		end
		love.graphics.setCaption("AStrat // Phase 6: conquering provinces")
		empires = maxemp
	end
end

function conquer()
	local current_province = mr(#provinces)
	while neighbours[current_province] == 0 do
		current_province = mr(#provinces)
	end
	local ns = {}
	for i, neighbour in ipairs(neighbours[current_province]) do
		local allegiance = provinces[neighbour]
		ns[allegiance] = (ns[allegiance] or 0) + 1
	end
	nneighbours = #neighbours[current_province]
	if nneighbours == 0 then
		return
	end
	new_owner = mr(nneighbours)
	for k, o in pairs(ns) do
		new_owner = new_owner - o
		if new_owner <= 0 then
			--print('change owner from', provinces[current_province], 'to', k)
			number_of_provinces[provinces[current_province]] = number_of_provinces[provinces[current_province]] - 1
			number_of_provinces[k] = number_of_provinces[k] + 1
			if number_of_provinces[provinces[current_province]] == 0 then
				empires = empires - 1
				--print('!!!', empires)
				if empires == 1 then
					complete_victory = true
					love.graphics.setCaption("AStrat // Phase 7: victory for empire "..colorname[k])
				end
			end
			provinces[current_province] = k
			break
		end
	end
	love.timer.sleep(100)
end

local LOOP = 0
local isDown = love.keyboard.isDown
function love.draw2()
	if not isDown' ' then
		if maxemp == 0 then
			if istep <= prec then
				love.graphics.setRenderTarget(fb)
			end
			love.graphics.push()
			love.graphics.scale(pix,pix)
			for i = 1, max, istep do
				for j = 1, max, istep do
					if type(map[i][j]) ~= 'number' then
						love.graphics.setColor(unpack(colors[map[i][j]]))
						love.graphics.rectangle('fill', i - 1, j - 1, istep, istep)
					end
				end
			end
			love.graphics.setColor(255,255,255)
			love.graphics.pop()
			if istep <= prec then
				love.graphics.setRenderTarget()
				love.graphics.draw(fb, 0, 0)
			end
		elseif not landfull then
			for n, pixl in ipairs(changed_pixels) do
				local i, j = unpack(pixl)
				love.graphics.setRenderTarget(altfb)
				love.graphics.draw(emp_fbs[map[i][j]], 0, 0)
				love.graphics.push()
				love.graphics.scale(pix,pix)
				love.graphics.rectangle('fill', i - 1, j - 1, istep, istep)
				love.graphics.setRenderTarget()
				love.graphics.pop()
				altfb, emp_fbs[map[i][j]] = emp_fbs[map[i][j]], altfb
			end
			changed_pixels = {}
			love.graphics.setColor(255,255,255)
			love.graphics.draw(fb, 0, 0)
			if mode == 0 then
				for i = 1, #emp_fbs do
					love.graphics.setColor(unpack(colors[i]))
					love.graphics.draw(emp_fbs[i], 0, 0)
				end
			elseif mode <= #emp_fbs then
				love.graphics.setColor(0,0,0)
				love.graphics.draw(emp_fbs[mode], 0, 0)
				love.graphics.setColor(255,255,255)
			end
		else
			for n, pixl in ipairs(changed_pixels) do
				local i, j = unpack(pixl)
				local m = pmap[i][j]
				province_fbs[m]:setPixel((i-1)/prec, (j-1)/prec, 255, 255, 255, 255)
			end
			if provincesfull and LOOP == 0 then
				LOOP = 1
				for m=1,#provinces do
					province_image[m] = love.graphics.newImage(province_fbs[m])
					province_image[m]:setFilter('nearest', 'nearest')
				end
			end
			changed_pixels = {}
			love.graphics.setColor(255,255,255)
			love.graphics.draw(fb, 0, 0)
			if not provincesfull then
				for i = 1, #emp_fbs do
					love.graphics.setColor(colors[i][1], colors[i][2], colors[i][3], 170)
					love.graphics.draw(emp_fbs[i], 0, 0)
				end
			end
			love.graphics.push()
			love.graphics.scale(pix*prec,pix*prec)
			if mode == 0 then
				for i = 1, #province_fbs do
					love.graphics.setColor(unpack(colors[provinces[i]]))
					love.graphics.draw(province_image[i], 0, 0)
				end
			elseif mode <= #province_fbs then
				love.graphics.setColor(0,0,0)
				love.graphics.draw(province_image[mode], 0, 0)
				love.graphics.setColor(255,255,255)
			end
			love.graphics.pop()
			if not provincesfull then
				love.graphics.rectangle('fill', 0, 506, 512 * pix_so_far / total_pix, 6)
			end
		end
	end
end

function love.keypressed2(key)
	if key == 'escape' then love.event.push'q' end
	if key == 'm' then
		local newv = (isDown'lshift' or isDown'rshift') and (mode-1) or (mode+1)
		if not landfull then
			mode = newv % (#emp_fbs+2)
		else
			mode = newv % (#province_fbs+2)
		end
	elseif key == 'n' then
		mode = 0
	end
end

local selx, sely = 0, 0
function love.update(dt)
	local x, y = love.mouse.getPosition()
	selx = math.floor(x / 512 * 3) 
	sely = math.floor(y / 512 * 2) 
end

function love.draw()
	love.graphics.setColor(100, 100, 255)
	love.graphics.rectangle('fill', 0, 0, 512, 512)
	love.graphics.setColor(180, 180, 255)
	love.graphics.rectangle('fill', selx * 512 / 3 + 10, sely * 512 / 2 + 10, 512 / 3 - 20, 512 / 2 - 20)
	love.graphics.setColor(0, 0, 0)
	for i = 0, 2 do
		for j = 0, 1 do
			love.graphics.print("Pixel size: " .. (2^i), i * 512 / 3 + 20, j * 512 / 2 + 20)
			love.graphics.print(j == 0 and "Block continent" or "Random continent", i * 512 / 3 + 20, j * 512 / 2 + 40)
		end
	end
	love.graphics.setColor(255,255,255)
end

function love.mousepressed(x, y, m)
	prec = 2^selx / 4
	FIELDS_PER_PROVINCE = 100 / prec
	xdir = {-prec, 0, prec, 0}
	ydir = {0, -prec, 0, prec}
	if sely == 0 then
		istep = istep / 2
	end
	for i = 1,max,istep do
		map[i] = {}
		for j = 1,max,istep do
			map[i][j] = 'w'
		end
	end
	if sely == 0 then
		for i = 1+istep,max-istep,istep do
			for j = 1+istep,max-istep,istep do
				map[i][j] = 'l'
			end
		end
	else
		for i = 1+istep,max-istep,istep do
			for j = 1+istep,max-istep,istep do
				if mr() < .75 then
					map[i][j] = 'l'
				end
			end
		end
	end
	love.update = love.update2
	love.draw = love.draw2
	love.mousepressed = nil
	love.keypressed = love.keypressed2
end

function love.keypressed(key)
	if key == 'escape' then love.event.push'q' end
end
